"""Coinigy API Facade"""

from marshmallow import fields, pre_load

from bors.app.log import LoggerMixin
from bors.api.requestor import Req
from bors.api.websock import SockMixin
from bors.generics.coinigy import NotificationSchema

from nombot.generics.request import RequestSchema
from nombot.generics.response import WSResponseSchema, ResponseSchema


class CoinigyResponseSchema(ResponseSchema):
    """Schema defining how the API will respond"""
    event = fields.Str()  # for WS response
    notifications = fields.List(fields.Nested(NotificationSchema()))
    err_num = fields.Str()
    err_msg = fields.Str()

    @pre_load
    def combine_errors(self, in_data):  # pylint: disable=no-self-use
        """Convert the error to the expected output"""
        if "err_num" in in_data:
            in_data["errors"] = dict()
            in_data["errors"][in_data["err_num"]] = in_data["err_msg"]

    def get_result(self, data):
        """Return the actual result data"""
        return data.get("data", "")

    class Meta:
        """Add 'data' field"""
        strict = True
        additional = ("data",)


class CoinigyWSResponseSchema(WSResponseSchema):
    """Schema defining the message type from a websocket"""
    def get_result(self, data):
        """Return the actual result data"""
        return data


class Coinigy(LoggerMixin, SockMixin):  # pylint: disable=R0902
    """
        This class implements coinigy's REST api as documented in the
        documentation available at:
        https://github.com/coinigy/api
    """
    name = "coinigy"

    ENDPOINT_OVERRIDES = {
        "history": "data",
        "asks": "data",
        "bids": "data",
        "orders": "data",
        "all": "data"
    }

    def __init__(self, context):
        """Launched by Api when we're ready to connect"""
        self.request_schema = RequestSchema
        self.result_schema = CoinigyResponseSchema
        self.ws_result_schema = CoinigyWSResponseSchema

        self.context = context

        # Websocket credentials object
        self.creds = self.context.get("credentials")

        # API credetials object
        payload = {
            'X-API-KEY': self.creds.get("apikey"),
            'X-API-SECRET': self.creds.get("secret")
        }

        self.create_logger()
        self.log.debug(f"Starting API Facade {self.name}")

        # Web request pool
        self.req = Req(self.context["conf"]["endpoints"]["rest"],
                       payload,
                       self.log)

        self.subscribed_chans = None

        # We don't need to deal directly with requests, so we pass them through
        self.call = self.req.call

    def shutdown(self):
        """Perform last-minute stuff"""
        self.log.info(f"Shutting down API interface instance for {self.name}")



"""Contexts for component integrations"""

from marshmallow import fields

from bors.generics.context import ApiContextSchema, StrategyContextSchema

from nombot.generics.config import ApiServiceConfSchema, ApiCredConfSchema


class NomApiContextSchema(ApiContextSchema):
    """Used for sharing information about a single API between instances"""
    conf = fields.Nested(ApiServiceConfSchema())  # API's config
    credentials = fields.List(fields.Nested(ApiCredConfSchema()))
    currencies = fields.List(fields.Str())  # List of currencies to monitor


class NomStrategyContextSchema(StrategyContextSchema):
    """Context to share information among strategies"""
    api_context = fields.Nested(NomApiContextSchema())  # API's context



"""
Build the context and pipeline; manage the API
"""
from bors.app.builder import AppBuilder

from nombot.generics.context import NomApiContextSchema
from nombot.generics.context import NomStrategyContextSchema


class NomAppBuilder(AppBuilder):
    """Class that assembles and runs the application"""
    api_context_schema = NomApiContextSchema
    strategy_context_schema = NomStrategyContextSchema

    def create_api_context(self, cls):
        """Create and return an API context"""
        return self.api_context_schema().load({
            "name": cls.name,
            "conf": self.conf.get_api_service(cls.name),
            "calls": self.conf.get_api_calls(),
            "currencies": self.conf.get_currencies(cls.name),
            "credentials": self.conf.get_api_credentials(cls.name),
            "log_level": self.conf.get_log_level(),
            "cls": cls,
            "shared": {},
            "inst": [],
            "callback": self.receive,
        })



"""Generic exchange schemas"""

from marshmallow import fields as f

from bors.generics.common import ResultSchema, RequestSchema, GenericSchema


class NotificationSchema(ResultSchema):
    """Notification!"""
    notification_id = f.Str(required=True)
    notification_type_title = f.Str(required=True)
    notification_type_message = f.Str(required=True)
    notification_style = f.Str(required=True)
    notification_vars = f.Str()
    notification_title_vars = f.Str()
    notification_pinned = f.Str()
    notification_sound = f.Str()
    notification_sound_id = f.Str()


class AccountSchema(ResultSchema):
    """Account with an exchange"""
    auth_id = f.Str(required=True)
    auth_nickname = f.Str()
    exch_id = f.Str(required=True)
    exch_name = f.Str(required=True)
    auth_active = f.Str()
    auth_trade = f.Str()
    auth_updated = f.Str()


class RefreshBalanceSchema(GenericSchema):
    """Used to refresh a balance on an account"""
    auth_id = f.Str(required=True)


class BalanceSchema(ResultSchema):
    """A balance for a given currency or auth_id"""
    balance_curr_code = f.Str(required=True)
    balance_amount_avail = f.Float()
    balance_amount_held = f.Float()
    balance_amount_total = f.Float()
    btc_balance = f.Float()
    last_price = f.Float()


class OrderSchema(ResultSchema):
    """Order"""
    exch_name = f.Str(required=True)
    mkt_name = f.Str(required=True)
    limit_price = f.Float(required=True)
    operator = f.Str()
    order_id = f.Int(required=True)
    order_type = f.Str(required=True)
    order_price_type = f.Str(required=True)
    order_status = f.Str(required=True)
    quantity = f.Float(required=True)
    order_time = f.Str(required=True)
    foreign_order_id = f.Str()
    auth_id = f.Str(required=True)
    auth_nickname = f.Str()
    quantity_remaining = f.Float()
    stop_price = f.Float()
    price_type_id = f.Str()
    exch_code = f.Str(required=True)
    display_name = f.Str()


class AllOrdersSchema(ResultSchema):
    """Result containing open orders and history"""
    open_orders = f.List(f.Nested(OrderSchema()))
    order_history = f.List(f.Nested(OrderSchema()))


class AlertSchema(ResultSchema):
    """Alert"""
    exch_name = f.Str(required=True)
    mkt_name = f.Str(required=True)
    price = f.Float(required=True)
    operator = f.Str()  # <,>
    alert_history_id = f.Str()
    timestamp = f.Str(required=True)
    alert_price = f.Float(required=True)
    alert_note = f.Str()
    operator_text = f.Str()
    display_name = f.Str()


class AllAlertsSchema(ResultSchema):
    """Result containing active alerts and history"""
    open_alerts = f.List(f.Nested(AlertSchema()))
    alert_history = f.List(f.Nested(AlertSchema()))


class FavoriteTickSchema(ResultSchema):
    """Tick item from favorites"""
    exch_id = f.Int(required=True)
    exch_code = f.Str(required=True)
    exch_name = f.Str(required=True)
    mkt_id = f.Int(required=True)
    exchmkt_id = f.Int(required=True)
    display_name = f.Str(required=True)
    mkt_name = f.Str(required=True)
    primary_curr = f.Str(required=True)
    base_curr = f.Str(required=True)
    last_price = f.Int(required=True)
    btc_volume_24 = f.Float(required=True)
    volume_24 = f.Float(required=True)


class NewsItemSchema(ResultSchema):
    """News item"""
    id = f.Str(required=True)  # pylint: disable=invalid-name
    url = f.Str(required=True)
    title = f.Str(required=True)
    pubDate = f.Str(required=True)
    timestamp = f.Str(required=True)
    feed_id = f.Int(required=True)
    published_date = f.Str(required=True)
    feed_name = f.Str(required=True)
    feed_url = f.Str(required=True)
    feed_enabled = f.Int(required=True)
    feed_description = f.Str(required=True)
    url_field = f.Str(required=True)
    title_field = f.Str(required=True)
    date_field = f.Str(required=True)
    feed_image = f.Str(required=True)


class OrderTypeSchema(ResultSchema):
    """Order Type"""
    order_type_id = f.Int(required=True)
    name = f.Str(required=True)


class PriceTypeSchema(ResultSchema):
    """Price Type"""
    price_type_id = f.Int(required=True)
    name = f.Str(required=True)


class OrderTypesCallSchema(ResultSchema):
    """Response containing order_types and price_types"""
    order_types = f.List(f.Nested(OrderTypeSchema()))
    price_types = f.List(f.Nested(PriceTypeSchema()))


class CreateAlertSchema(GenericSchema):
    """Schema to generate a new alert"""
    exch_code = f.Str(required=True)  # BITF
    market_name = f.Str(required=True)  # BTC/USD
    alert_price = f.Float(required=True)
    alert_note = f.Str()


class AlertReferenceSchema(ResultSchema):
    """Schema to reference an alert during CRUD operations"""
    alert_id = f.Int(required=True)


class CreateOrderSchema(GenericSchema):
    """Schema to generate a new order"""
    auth_id = f.Int(required=True)
    exch_id = f.Int(required=True)
    mkt_id = f.Int(required=True)
    order_type_id = f.Int(required=True)
    price_type_id = f.Int(required=True)
    limit_price = f.Float(required=True)
    order_quantity = f.Float(required=True)


class OrderReferenceSchema(ResultSchema):
    """Schema to reference an order during CRUD operations"""
    internal_order_id = f.Int(required=True)


class ExchangeSchema(ResultSchema):
    """An exchange"""
    exch_id = f.Int(required=True)
    exch_name = f.Str(required=True)
    exch_code = f.Str(required=True)
    exch_fee = f.Float(required=True)
    exch_trade_enabled = f.Int(required=True)
    exch_balance_enabled = f.Int(required=True)
    exch_url = f.Str(required=True)


class ExchangeReferenceSchema(GenericSchema):
    """Used to request market data"""
    exchange_code = f.Str(required=True)


class MarketSchema(ResultSchema):
    """A currency trade pair"""
    exch_id = f.Int(required=True)
    exch_name = f.Str(required=True)
    exch_code = f.Str(required=True)
    mkt_id = f.Int(required=True)
    mkt_name = f.Str(required=True)
    exchmkt_id = f.Int(required=True)


class MarketDataRequestSchema(RequestSchema):
    """Request object for market data"""
    exchange_code = f.Str(required=True)
    exchange_market = f.Str(required=True)
    type = f.Str(required=True)


class MarketDataItemSchema(ResultSchema):
    """A single item of market data"""
    price = f.Float(required=True)
    quantity = f.Float(required=True)


class OrderItemSchema(MarketDataItemSchema):
    """A market data item that represents an ask or bid"""
    total = f.Float(required=True)


class HistoryItemSchema(MarketDataItemSchema):
    """A single order history item"""
    time_local = f.Str(required=True)
    type = f.Str(required=True)


class MarketDataSchema(ResultSchema):
    """A result encompassing some form of market data"""
    exch_code = f.Str(required=True)
    primary_curr_code = f.Str(required=True)
    secondary_curr_code = f.Str(required=True)
    type = f.Str(required=True)


class HistorySchema(MarketDataSchema):
    """A history of market data"""
    history = f.List(f.Nested(HistoryItemSchema()), required=True)


class AsksSchema(MarketDataSchema):
    """A list of open sell market orders"""
    asks = f.List(f.Nested(OrderItemSchema()), required=True)


class BidsSchema(MarketDataSchema):
    """A list of open buy market orders"""
    bids = f.List(f.Nested(OrderItemSchema()), required=True)


class OrdersSchema(MarketDataSchema):
    """A list of open buy AND sell market orders"""
    asks = f.List(f.Nested(OrderItemSchema()), required=True)
    bids = f.List(f.Nested(OrderItemSchema()), required=True)


class AllMarketDataSchema(MarketDataSchema):
    """A list of open buy and sell market orders, and the purchase history"""
    history = f.List(f.Nested(HistoryItemSchema()), required=True)
    asks = f.List(f.Nested(OrderItemSchema()), required=True)
    bids = f.List(f.Nested(OrderItemSchema()), required=True)


class TickerRequestSchema(RequestSchema):
    """Request object for ticker data"""
    exchange_code = f.Str(required=True)
    exchange_market = f.Str(required=True)


class TickSchema(ResultSchema):
    """An item from a ticker"""
    exchange = f.Str(required=True)
    market = f.Str(required=True)
    last_trade = f.Float(required=True)
    high_trade = f.Float(required=True)
    low_trade = f.Float(required=True)
    current_volume = f.Float(required=True)
    timestamp = f.Str(required=True)
    ask = f.Float(required=True)
    bid = f.Float(required=True)


class WsTradeChannel(ResultSchema):
    """An item from a websocket trade channel"""
    market_history_id = f.Int(required=True)
    channel = f.Str(required=True)
    exchange = f.Str(required=True)
    marketid = f.Int(required=True)
    label = f.Str(required=True)
    tradeid = f.Str(required=True)
    price = f.Float(required=True)
    quantity = f.Float(required=True)
    total = f.Float(required=True)
    type = f.Str(required=True)
    exchId = f.Int(required=True)
    time = f.Str(required=True)
    timestamp = f.Str(required=True)
    time_local = f.Str(required=True)


class WsOrderChannel(ResultSchema):
    """An item from a websocket order channel"""
    exchange = f.Str(required=True)
    label = f.Str(required=True)
    ordertype = f.Str(required=True)
    price = f.Float(required=True)
    quantity = f.Float(required=True)
    total = f.Float(required=True)
    timestamp = f.Str(required=True)



"""Trading strategies"""
import pandas as pd

from bors.app.strategy import IStrategy


class OHLCV:
    """Implementation of OHLCV data (open, high, low, close, volume)"""
    history = pd.DataFrame()

    def __init__(self, hist_size):
        self.hist_size = hist_size

    def update(self, data):
        """Update the data using the latest information"""
        pass

    def dump(self):
        """Dump the data for the strategy pipeline"""
        pass


class OHLCVStrategy(IStrategy):
    """Strategy to supplement/act upon incoming data"""
    name = "ohlc_strategy"

    def __init__(self, hist_size=100):
        self._data = OHLCV(hist_size)
        self.hist_size = hist_size

    def bind(self, context):
        """Bind actions to the strategy context for a given result"""
        self._data.update(context)

        # initialze supplemented data
        context["strategy"].update({
            "ohlc_hist_size": self.hist_size,
            "ohlc": self._data.dump(),
            })

        return context



"""
Supplementary trading strategy
"""

from bors.app.strategy import IStrategy


class StockSupplement(IStrategy):
    """Print Strategy implementation"""
    history = StockDataFrame()

    def bind(self, context):
        """
        Bind the strategy to the middleware pipeline,
        returning the context
        """
        self.supplement(context)

        # just pass-through
        return context

    def supplement(self, context):
        """Supplement the context with stockstats"""
        self.history.append(self.parse(context.get('result')))

    def parse(self, result):
        """Parse the results to generate stock data as exchange->src->dest"""
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test the `bors` configuration overrides"""


import unittest
from copy import deepcopy

from nombot.app.config import NomAppConf
from nombot.generics.config import NomConfSchema


DEFAULT_TEST_CONFIG = {
    "currencies": ["BTC", "USD", "USDT"],
    "api": {
        "ratelimit": 1,
        "services": [
            {
                "name": "test_service",
                "currencies": ["BTC", "ETH", "XRP"],
                "credentials": [
                    {
                        "name": "test_api_name",
                        "apiKey": "test_key",
                        "secret": "test_secret",
                    },
                ],
                "subscriptions": {"TICKER_X": "ticker"},
                "exchanges": ["bittrex", "coinbase"],
                "endpoints": {
                    "rest": "rest_url",
                    "websocket": "sock_url",
                },
            },
        ],
        "calls": {"test_call1": {"AAA": 123}},
    },
    "log_level": "INFO",
}


class TestConfig(unittest.TestCase):
    """Tests for nombot configuration parser"""

    def setUp(self):
        """Set up test fixtures, if any"""
        self.config = deepcopy(DEFAULT_TEST_CONFIG)

    def test_schema(self):
        """Test the schema definition"""
        l_config = deepcopy(self.config)
        del l_config["api"]["ratelimit"]  # bors field unused by nombot
        config = NomConfSchema().load(l_config).data["conf"]
        self.assertDictEqual(config, l_config)

    def test_config_load(self):
        """Test the load of a configuration"""
        config = NomAppConf(self.config).raw_conf
        self.assertDictEqual(config, self.config)


class TestCurrencies(unittest.TestCase):
    """Tests for nombot currency parser"""

    def setUp(self):
        """Set up test fixtures, if any"""
        self.config = deepcopy(DEFAULT_TEST_CONFIG)

    def test_get_currencies(self):
        """Test the schema definition"""
        config = NomAppConf(self.config)
        self.assertListEqual(config.get_currencies(),
                             self.config["currencies"])

    def test_get_service_currencies(self):
        """Test the schema definition"""
        config = NomAppConf(self.config)

        expected_curr = self.config["api"]["services"][0]["currencies"] + \
            self.config["currencies"]
        result_curr = config.get_currencies("test_service")

        self.assertListEqual(expected_curr, result_curr)

    def test_global_without_service(self):
        """If no service currencies, return only global"""
        l_config = deepcopy(self.config)
        del l_config["api"]["services"][0]["currencies"]
        config = NomAppConf(l_config)

        expected_curr = self.config["currencies"]
        result_curr = config.get_currencies("test_service")

        self.assertListEqual(expected_curr, result_curr)

    def test_global_with_only_service(self):
        """If no global currencies, return only service"""
        l_config = deepcopy(self.config)
        del l_config["currencies"]
        config = NomAppConf(l_config)

        expected_curr = self.config["api"]["services"][0]["currencies"]
        result_curr = config.get_currencies("test_service")

        self.assertListEqual(expected_curr, result_curr)

    def test_no_currencies(self):
        """If no currencies, return empty list"""
        l_config = deepcopy(self.config)
        del l_config["currencies"]
        del l_config["api"]["services"][0]["currencies"]
        config = NomAppConf(l_config)

        expected_curr = []  # type: list
        result_curr = config.get_currencies("test_service")

        self.assertListEqual(expected_curr, result_curr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test the `bors` configuration overrides"""


import unittest

from nombot.app.config import NomAppConf
from nombot.generics.config import NomConfSchema


DEFAULT_TEST_CONFIG = {
    "currencies": ["BTC", "USD", "USDT"],
    "api": {
        "ratelimit": 1,
        "calls": {"test_call1": {"AAA": 123}},
        "services": [
            {
                "name": "test_service",
                "currencies": ["BTC", "ETH", "XRP"],
                "credentials": [
                    {
                        "name": "test_api_name",
                        "apiKey": "test_key",
                        "secret": "test_secret",
                    },
                ],
                "subscriptions": {"TICKER_X": "ticker"},
                "exchanges": ["bittrex", "coinbase"],
                "endpoints": {
                    "rest": "rest_url",
                    "websocket": "sock_url",
                },
            },
        ]
    },
    "log_level": "INFO",
}


class TestConfig(unittest.TestCase):
    """Tests for nombot configuration parser"""

    def setUp(self):
        """Set up test fixtures, if any"""
        self.config = DEFAULT_TEST_CONFIG.copy()

    def tearDown(self):
        """Tear down test fixtures, if any"""

    def test_schema(self):
        """Test the schema definition"""
        assert NomConfSchema().load(self.config).data["conf"] == self.config

    def test_config_load(self):
        """Test the schema definition"""
        assert NomAppConf(self.config).raw_conf == self.config



"""Configuration module"""

from nombot.generics.config import NomConfSchema
from bors.app.config import AppConf


class NomAppConf(AppConf):
    """NomBot configuration Object"""
    schema = NomConfSchema

    def get_currencies(self, service=None):
        """Returns the currencies that we'll be working with"""
        try:
            self.get_api_services_by_name()
            if service is not None:
                svc_conf = self.services_by_name[service]
                return svc_conf.get("currencies", []).copy() \
                    + self.conf.get("currencies", []).copy()
            return self.conf.get("currencies", []).copy()
        except AttributeError:
            return []



"""
Generic API Interface, mixin, and request maps/types
"""

from marshmallow import fields, Schema, post_load
from nombot.api.request import Request, REQUEST_MAP


class RequestSchema(Schema):
    """Schema defining the data structure the API can be called with"""
    callname = fields.Str(required=True)
    payload = fields.Dict()

    @post_load
    def make_request(self, data):
        """Parse the outgoing schema"""
        callname = self.context.get("callname")
        try:
            payload = REQUEST_MAP[callname].dump(data)  # type: ignore
        except AttributeError:
            payload = None

        request = {
            "callname": callname,
            "payload": payload
        }
        return Request(**request)

    class Meta:
        """Stricty"""
        strict = True



"""CCXT API Facade"""

import asyncio
from itertools import product
from dataclasses import dataclass

import ccxt.async as ccxt
from ccxt.base.errors import ExchangeNotAvailable, ExchangeError

from marshmallow import pre_load

from bors.app.log import LoggerMixin

from nombot.generics.request import RequestSchema
from nombot.generics.response import ResponseSchema


class CCXTResponseSchema(ResponseSchema):
    """Generic response class"""
    @pre_load
    def prep_response(self, in_data):
        """Set the response to the correct output"""
        in_data["result"] = in_data


@dataclass
class CCXTExchange:
    """Exchange data object"""
    name: str
    currencies: list
    rate_limit: int = None
    credentials: dict = None

    _ex = None
    loop = asyncio.get_event_loop()

    avail_currencies = None
    _currencies = None

    avail_markets = None
    markets = None

    avail_symbols = None
    symbols = None

    def __post_init__(self):
        # instantiate exchange object
        self._ex = getattr(ccxt, self.name)()

        if self.rate_limit is not None:
            self._ex.rate_limit = self.rate_limit
        self._ex.enable_rate_limit = True

        if self.credentials is not None:
            self._ex.apiKey = self.credentials.get("apiKey", None)
            self._ex.secret = self.credentials.get("secret", None)

        self.loop.run_until_complete(self.load())

    def has(self, callname):
        """Returns `has` value of exchange"""
        return self._ex.has.get(callname, False)

    async def load(self, reload=False, *args, **kwargs):
        """Load the markets; populating the exchange object with data"""
        if self.markets is not None and not reload:
            return

        self.avail_markets = await self._ex.load_markets(*args, **kwargs)

        self.avail_currencies = getattr(self._ex, "currencies", {})
        if not self.currencies:
            # copy all currencies
            self._currencies = self.avail_currencies
        else:
            # copy relevant currencies
            self._currencies = {
                curr: self.avail_currencies[curr]
                for curr in self.currencies
                if curr in self.avail_currencies
            }

        self.avail_symbols = getattr(self._ex, "symbols", [])
        currencies = self._currencies.keys()
        self.symbols = [
            "/".join(pair)
            for pair in product(currencies, currencies)
            if "/".join(pair) in self.avail_symbols
        ]

        self.markets = {
            sym: self.avail_markets[sym]
            for sym in self.symbols
        }

    async def call_over_syms(self, callname, *args, **kwargs):
        """Cycle through configured exchanges and symbols and make a call"""
        results = {}
        for sym in self.markets.keys():
            try:
                results[sym] = await self.call(callname, sym, *args, **kwargs)
            except (ExchangeNotAvailable, ExchangeError):
                pass
        return results

    async def call(self, callname, *args, **kwargs):
        """Generalized async `call` method, pass callname and parameters"""
        try:
            return await getattr(self._ex, callname)(*args, **kwargs)
        except TypeError:
            raise AttributeError(f"Failed to execute call {callname} on "
                                 f"exchange {self._ex.name}")

    async def close(self):
        """Close all exchange connections"""
        await self._ex.close()

    def shutdown(self):
        """Teardown the aio loop"""
        self.loop.run_until_complete(self.close())
        self.loop.close()


class CCXT:
    """CCXTExchange wrapper"""
    _ex = {}  # type: dict

    def __init__(self, log, conf, context):
        self.log = log
        self.conf = conf

        rate_limit = self.conf.get("rate_limit", None)
        currencies = self.conf.get("currencies", None)
        exchanges = self.conf.get("exchanges", None)
        credentials = context.get("credentials", None)

        if exchanges is None or not exchanges:
            exchanges = ccxt.exchanges

        # Instantiate exchange objects
        for exch in exchanges:

            # Extract credentials if they exist
            creds = None
            if credentials is not None:
                for cred in credentials:
                    if cred.get("name") == exch:
                        creds = cred.copy()

            # launch exchange
            self._ex[exch] = \
                CCXTExchange(exch, currencies=currencies,
                             rate_limit=rate_limit, credentials=creds)

    def call_capable(self, exch, callname):
        """Determine if the exchange supports the call"""
        if self._ex[exch].has(callname):
            return self._ex[exch].has(callname)
        self.log.warn(f"No method available -- "
                      f"exchange: {exch}; "
                      f"call: {callname}")
        return False

    def call_on_exchanges(self, calltype, callname, *args, **kwargs):
        """Cycle through all configured exchanges to make a call"""
        results = {}
        for ex in self._ex.values():
            if self.call_capable(ex.name, callname):
                try:
                    results[ex.name] = ex.loop.run_until_complete(
                        getattr(ex, calltype)(callname, *args, **kwargs))
                except (ExchangeNotAvailable, ExchangeError):
                    pass
        return results

    def shutdown(self):
        """Shutdown / cleanup"""
        for ex in self._ex.values():
            ex.shutdown()


class CCXTApi(LoggerMixin):  # pylint: disable=R0902
    """
        This class implements ccxt's REST api as documented in the
        documentation available at:
        https://github.com/ccxt/ccxt/wiki/Manual
    """
    name = "ccxt"

    local_overrides = {
        "fetchOHLCV": "call_over_syms",
        "fetchOrderBook": "call_over_syms",
        "fetchTicker": "call_over_syms",
        "fetchTrades": "call_over_syms",
    }

    def __init__(self, context):
        """Launched by Api when we're ready to connect"""
        self.request_schema = RequestSchema
        self.result_schema = CCXTResponseSchema

        self.context = context
        self.conf = context.get("conf")

        self.create_logger()
        self.log.debug(f"Starting API Facade {self.name}")

        self.ccxt = CCXT(self.log, self.conf, self.context)

    def call(self, callname, *args, **kwargs):
        """Substitute for REST api as defined in bors.api.requestor.Req"""
        return self.ccxt.call_on_exchanges(
            self.local_overrides.get(callname, "call"),
            callname, *args, **kwargs)

    def shutdown(self):
        """Perform last-minute stuff"""
        self.log.info(f"Shutting down API interface instance for {self.name}")

        # Take care of any currently running tasks in open loops
        for task in asyncio.Task.all_tasks():
            task.cancel()

        self.ccxt.shutdown()



"""
Generic API Interface, mixin, and response maps/types
"""

from bors.common.dotobj import DotObj
from nombot.generics import exchange as X


RESPONSE_MAP = {
    # commented, as they are not implemented
    "fetchBalance": X.BalanceSchema(many=True),
    "fetchMarkets": X.MarketSchema(),
    # "fetchOHLCV": X.OHLCVSchema(many=True),
    "fetchOrderBook": X.OrderBookSchema(many=True),
    "fetchOrders": X.OrderSchema(many=True),
    "fetchOpenOrders": X.OrderSchema(many=True),
    "fetchClosedOrders": X.OrderSchema(many=True),
    "fetchTicker": X.TickerSchema(),
    "fetchTickers": X.TickerSchema(many=True),
    "fetchTrades": X.TradeSchema(many=True),
    "fetchMyTrades": X.MyTradeSchema(many=True),
}


class Result(DotObj):
    """A bare result object"""
    def __init__(self, **kwargs):
        self.callname = kwargs.get('callname', None)
        self.channel = kwargs.get('channel', None)
        self.response_type = kwargs.get('response_type', None)
        self.result = kwargs.get('results', None)
        self.errors = kwargs.get('errors', None)



"""
Generic API Interface, mixin, and request maps/types
"""

from bors.generics import coinigy as X


REQUEST_MAP = {
    "refreshBalance": X.RefreshBalanceSchema(),
    "addAlert": X.CreateAlertSchema(),
    "deleteAlert": X.AlertReferenceSchema(),
    "addOrder": X.CreateOrderSchema(),
    "cancelOrder": X.OrderReferenceSchema(),
    "markets": X.ExchangeReferenceSchema(),
    "data": X.MarketDataRequestSchema(),
    "ticker": X.TickerRequestSchema(),
}


class Request:
    """A bare request object"""
    def __init__(self, **kwargs):
        self.callname = kwargs.get('callname', None)
        self.payload = kwargs.get('payload', None)



"""
Simple Market Maker strategy
"""

from bors.app.strategy import IStrategy
from bors.algorithms.echo import echo


class MarketMaker(IStrategy):
    """Market Maker Strategy imolementation"""
    def bind(self, context):
        """
        Bind the strategy to the middleware pipeline,
        returning the context
        """
        echo("ALGO!")
        # currently, just a pass-through
        return context



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `nombot` package."""


import unittest

#from trader import main


class TestNombot(unittest.TestCase):
    """Tests for `nombot` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_main(self):
        """Test main trader"""



"""
Generic API Interface, mixin, and response maps/types
"""

from bors.common.dotobj import DotObj
from bors.generics import coinigy as X


RESPONSE_MAP = {
    "accounts": X.AccountSchema(many=True),
    "balances": X.BalanceSchema(many=True),
    "alerts": X.AllAlertsSchema(),
    "newsFeed": X.NewsItemSchema(many=True),
    "orderTypes": X.OrderTypesCallSchema(),
    "refreshBalance": X.BalanceSchema(),
    "addOrder": X.OrderReferenceSchema(),
    "exchanges": X.ExchangeSchema(many=True),
    "markets": X.MarketSchema(many=True),
    "data": X.AllMarketDataSchema(),
    "ticker": X.TickSchema(),
    "trade": X.WsTradeChannel(),
    "orders": X.WsOrderChannel(many=True),
    "Favorite": X.FavoriteTickSchema(many=True),
    "all": X.AllMarketDataSchema(),
}


class Result(DotObj):
    """A bare result object"""
    def __init__(self, **kwargs):
        self.callname = kwargs.get('callname', None)
        self.channel = kwargs.get('channel', None)
        self.response_type = kwargs.get('response_type', None)
        self.result = kwargs.get('result', None)
        self.errors = kwargs.get('errors', None)
        super().__init__(**kwargs)



"""Configuration definition"""

from marshmallow import Schema, fields, post_load

from bors.generics.config import ConfSchema


class ApiEndpointConfSchema(Schema):
    """API endpoint configuration object"""
    rest = fields.Str()
    websocket = fields.Str()


class ApiCredConfSchema(Schema):
    """API credential configuration object"""
    name = fields.Str()
    apiKey = fields.Str()
    secret = fields.Str()


class ApiServiceConfSchema(Schema):
    """API service configuration object"""
    name = fields.Str(required=True)
    currencies = fields.List(fields.Str())
    credentials = fields.List(fields.Nested(ApiCredConfSchema))
    subscriptions = fields.Dict()
    exchanges = fields.List(fields.Str())
    endpoints = fields.Nested(ApiEndpointConfSchema())


class ApiConfSchema(Schema):
    """API configuration object"""
    calls = fields.Dict()
    services = fields.List(fields.Nested(ApiServiceConfSchema()))


class NomConfSchema(ConfSchema):
    """Root configuration schema"""
    currencies = fields.List(fields.Str())
    api = fields.Nested(ApiConfSchema())



"""Generic exchange schemas"""

from marshmallow import fields as f
from marshmallow import pre_load

from bors.generics.common import ResultSchema


class ExchangeSchema(ResultSchema):
    """A generic schema for exchange meta-interfaces"""
    exchange = f.Str(required=True)

    def prepare(self, data):
        _data = []  # type: list
        for exch, result in data["result"].items():
            if exch != "result":
                _result = {}
                _result["exchange"] = exch
                for key, item in result.items():
                    if key != "result":
                        _result[key] = item
                _data.append(_result)
        return _data


class LimitSchema(ExchangeSchema):
    """A Limit"""
    min = f.Float()
    max = f.Float()


class AmountPriceSchema(ExchangeSchema):
    """Limits"""
    amount = f.Nested(LimitSchema, required=True)
    price = f.Nested(LimitSchema, required=True)


class FeeSchema(ExchangeSchema):
    """A fee"""
    currency = f.Str(required=True)
    cost = f.Float(required=True)
    rate = f.Float(required=True)


class OrderSchema(ExchangeSchema):
    """Order"""
    id = f.Str(required=True)
    datetime = f.Str(required=True)
    timestamp = f.Int(required=True)
    lastTradeTimestamp = f.Int(required=True)
    status = f.Str(required=True)
    symbol = f.Str(required=True)
    type = f.Str(required=True)
    side = f.Str(required=True)
    price = f.Float(required=True)
    amount = f.Float(required=True)
    filled = f.Float(required=True)
    remaining = f.Float(required=True)
    cost = f.Float(required=True)
    trades = f.List(f.List(f.Float()))  # [  [ price, amount ], ... ]
    fee = f.Nested(FeeSchema, required=True)
    info = f.Dict(required=True)

    def prepare(self, data):
        _data = []  # type: list
        for exch, result in data["result"].items():
            print(f"""EXCH: {exch}; RESULT: {result}""")
            if exch != "result":
                _result = {}
                _result["exchange"] = exch
                for key, item in result.items():
                    if key != "result":
                        _result[key] = item
                _data.append(_result)
        return _data


class OrderBookSchema(ExchangeSchema):
    """Order Book"""
    bids = f.List(f.List(f.Float()))  # [  [ price, amount ], ... ]
    asks = f.List(f.List(f.Float()))  # [  [ price, amount ], ... ]
    timestamp = f.Int()
    datetime = f.Str()
    nonce = f.Int()
    market = f.Str()

    def prepare(self, data):
        data = super().prepare(data)
        _data = []
        for result in data:
            for market, orderbook in result.items():
                if market != "exchange":
                    _result = {"exchange": result["exchange"]}
                    _result["market"] = market
                    for key, item in orderbook.items():
                        if key != "market":
                            _result[key] = item
                    _data.append(_result)
        return _data


class MarketSchema(ExchangeSchema):
    """Market"""
    active = f.Bool(required=True)
    symbol = f.Str(required=True)  # 'BTC/USD'
    base = f.Str(required=True)  # 'BTC'
    baseId = f.Str(required=True)
    quote = f.Str(required=True)  # 'USD'
    quoteId = f.Str(required=True)
    fee_loaded = f.Bool()
    id = f.Str(required=True)
    info = f.Dict(required=True)
    limits = f.Nested(AmountPriceSchema)
    maker = f.Float()
    taker = f.Float()
    percentage = f.Bool()
    precision = f.Nested(AmountPriceSchema)
    tierBased = f.Bool()


class TickerSchema(ExchangeSchema):
    """Ticker!"""
    symbol = f.Str(required=True)
    info = f.Dict()
    timestamp = f.Int()
    datetime = f.Str()
    high = f.Float()
    low = f.Float()
    bid = f.Float()
    bidVolume = f.Float()
    ask = f.Float()
    askVolume = f.Float()
    vwap = f.Float()
    open = f.Float()
    close = f.Float()
    last = f.Float()
    previousClose = f.Float()
    change = f.Float()
    percentage = f.Float()
    average = f.Float()
    baseVolume = f.Float()
    quoteVolume = f.Float()

    def prepare(self, data):
        _data = []
        for result in super().prepare(data):
            exchange = result["exchange"]
            del result["exchange"]
            for tick in result.values():
                _result = tick
                _result["exchange"] = exchange
                _data.append(_result)
        return _data


class TradeSchema(ExchangeSchema):
    """A single trade"""
    info = f.Dict(required=True)
    id = f.Str(required=True)
    timestamp = f.Int(required=True)
    datetime = f.Str(required=True)
    symbol = f.Str(required=True)
    order = f.Str()
    type = f.Str(required=True)
    side = f.Str(required=True)
    price = f.Float(required=True)
    amount = f.Float(required=True)

    def prepare(self, data):
        _data = []
        for result in super().prepare(data):
            exchange = result["exchange"]
            del result["exchange"]
            for market in result.values():
                for trade in market:
                    _result = trade
                    _result["exchange"] = exchange
                    _data.append(_result)
        return _data


class MyTradeSchema(ExchangeSchema):
    """A single trade on my account"""
    id = f.Str(required=True)
    timestamp = f.Int(required=True)
    datetime = f.Str(required=True)
    symbol = f.Str(required=True)
    order = f.Str()
    type = f.Str(required=True)
    side = f.Str(required=True)
    price = f.Float(required=True)
    amount = f.Float(required=True)
    cost = f.Float(required=True)
    fee = f.Nested(FeeSchema, required=True)
    info = f.Dict(required=True)


class BalanceSchema(ExchangeSchema):
    """The Balances"""
    info = f.List(f.Dict(), required=True)
    free = f.Dict()
    used = f.Dict()
    total = f.Dict()
    by_sym = f.Dict()

    def prepare(self, data):
        _data = []  # type: list
        for exch, result in data["result"].items():
            if exch != "result":
                _result = {"exchange": exch, "by_sym": {}}  # type: dict
                for key, item in result.items():
                    if key in ["info", "free", "used", "total"]:
                        _result[key] = item
                    elif key != "result":
                        _result["by_sym"][key] = item
                _data.append(_result)
        return _data



"""
Generic API Interface, mixin, and request maps/types
"""

from nombot.generics import exchange as X


REQUEST_MAP = {}


class Request:
    """A bare request object"""
    def __init__(self, **kwargs):
        self.callname = kwargs.get('callname', None)
        self.payload = kwargs.get('payload', None)



"""
Generic response types
"""

from marshmallow import fields, Schema, post_load, pre_load
from nombot.api.response import Result, RESPONSE_MAP


class DefaultSchema(Schema):
    """Default schema parser"""
    #MessageType = fields.Str(required=True)

    @post_load
    def generate_obj(self, data):  # pylint: disable=no-self-use
        """Generate new schema based on message type"""
        return RESPONSE_MAP[data["MessageType"]].dump(data["Data"])

    class Meta:
        """Data field is dyanamic type"""
        additional = ("Data",)


RESPONSE_MAP["default"] = DefaultSchema()


class CommonResponseSchema(Schema):
    """Common response schema"""
    errors = fields.Dict()

    def get_results(self, callname, data):  # pylint: disable=no-self-use
        """
        Retrieve the results from the parsed object
          ~~ Override this to match your API. ~~
        """
        # perform type-specific preperation of data, if it exists
        prep = getattr(RESPONSE_MAP[callname], "prepare", None)
        if callable(prep):
            data = prep(data)
        return data


class ResponseSchema(CommonResponseSchema):
    """Schema defining the data structure the API will respond with"""
    @post_load
    def populate_data(self, data):
        """Parse the incoming schema"""
        if "errors" in data:
            return Result(errors=data["errors"])
        callname = self.context.get("callname")
        results = {
            "callname": callname,
            "results": RESPONSE_MAP[callname]  # type: ignore
                .dump(self.get_results(callname, data))  # NOQA
        }
        return Result(**results)

    class Meta:
        """Stricty"""
        strict = True
        additional = ("result",)


class WSResponseSchema(CommonResponseSchema):
    """
    Schema defining the data structure from published messages on the websock
    """
    channel = fields.Str()

    @pre_load
    def prep_data(self, data):
        """Prepare the data for ingestion"""
        # pylint: disable=E1101
        res_type = self.context.get("response_type")
        try:
            sch = RESPONSE_MAP[res_type]
        except KeyError:
            sch = RESPONSE_MAP["default"]

        self.context["result"] = sch.dump(self.get_result(callname, data)).data  # NOQA
        return self.context

    @post_load
    def populate_data(self, data):
        """Parse the incoming schema"""
        if "errors" in data:
            return Result(errors=data["errors"])
        return Result(**self.context)



"""Pull in everything for export"""

from os.path import dirname, basename, isfile
import glob

MODULES = glob.glob(dirname(__file__)+"/*.py")

__all__ = [
    basename(f)[:-3] for f in MODULES
    if isfile(f) and not f.endswith('__init__.py')
    ]



"""Coinigy strategy, subscribes to all favorited channels"""
from bors.app.strategy import IStrategy
from bors.app.log import LoggerMixin

from bors.common.factory import Creator, Product
from bors.common.dotobj import DotObj


class CoinigyFacade(LoggerMixin):
    """Encapsulates some API functionality, initialized on result"""
    name = "coinigy_facade"
    chan_callback = None
    conf = None

    def __init__(self, inst, api_context):
        self.context = api_context
        self.conf = api_context.get("conf")
        self.api = inst

        self.create_logger()

    def get_channels(self, callback):
        """Setup possible channels"""
        self.chan_callback = callback
        self.api.wscall("channels", None, self._connect_channels)

    def _connect_channels(self, ename, error, data):   # pylint: disable=W0613
        """
        Dynamically generate the websocket channels based on exchange and
        currency configurations and what the server reports available.
        """
        if error:
            self.log.error(f"Channel request error: {error}")
            return

        # We've reached this point if we have a list of channels
        possible_channels = [item["channel"] for item in data[0]]

        # Assemble a list of configured channels
        channels = {chan: True for chan in
                    self.conf.get("subscriptions").keys()}

        # Assemble a list of possibilities based on the config exch/currency
        ecc = []
        for exch in self.conf["exchanges"]:
            for curr1 in self.context["currencies"]:
                for curr2 in self.context["currencies"]:
                    if curr1 != curr2:
                        ecc.append((exch, curr1, curr2))

        for exch, curr1, curr2 in ecc:
            for ortra in ["order", "trade"]:
                for chan in [
                        f"{ortra}-{exch}--{curr1}--{curr2}".upper(),
                        f"{ortra}-{exch}--{curr2}--{curr1}".upper()
                ]:
                    if chan in possible_channels:
                        if chan not in channels:
                            channels[chan] = False

        self.context["shared"]["channels"] = channels

        # Subscribe to channels that haven't been subscribed to yet
        chan_resp = {}
        for chan in [k for k, v in channels.items() if not v]:
            restype = None
            if chan.startswith("ORDER"):
                restype = "orders"
            elif chan.startswith("TRADE"):
                restype = "trade"
            if restype is not None:
                chan_resp[chan] = restype
                self.log.info(f"""CONNECTING CHANNEL!{chan}""")

        # Connect to channels
        self.api.add_channels(chan_resp)
        self.chan_callback(possible_channels)


class CoinigyResultParserFactory(Creator):
    """Generator of parser objects"""
    def _factory_method(self, *args, **kwargs):
        return CoinigyParser(*args, **kwargs)


class CoinigyParser(Product):
    """Result parsing class"""
    _call_lookup = {}  # type: dict

    def __init__(self, strategy_data, context):
        self.strategy_data = strategy_data
        self.context = context
        self.result = self.context.get("result").data
        self.api_ctx = self.context.get("conf")
        self.api_ctxs = self.context.get("api_contexts")

    def _lookup(self, callname):
        """Return a callable function based on the name"""
        return self._call_lookup.get(callname, self._default_parser)

    def _default_parser(self):
        """Return a result object for supplementation"""
        update = {}
        if self.result.callname is not None:
            update[self.result.callname] = self.result
        elif self.result.response_type is not None:
            update[self.result.response_type] = self.result
        elif self.result.channel is not None:
            update[self.result.channel] = self.result

        return self.strategy_data.coinigy.update({"data": update})

    def interface(self):
        """Call the update to the context based on the result received"""
        if self.result.callname is not None:
            self._lookup(self.result.callname)()
        elif self.result.channel is not None:
            self._lookup(self.result.channel)()


class CoinigyStrategyData(DotObj):
    """Object to hold all strategy data"""
    coinigy = {
        "apiname": "coinigy",
        "data": {}
    }

    def __str__(self):
        return self.coinigy.__str__()


class CoinigyStrategy(IStrategy):
    """Strategy to supplement/act upon Coinigy API events"""
    name = "coinigy_strategy"
    possible_channels = {}  # type: dict
    api_facade = None
    ws_facade = None

    def __init__(self):
        self._strategy_data = CoinigyStrategyData()

    def bind(self, context):
        """Bind actions to the strategy context for a given result"""
        if self.ws_facade is None or self.api_facade is None:
            api_context = context.get("api_context")
            for inst in api_context.get("inst"):
                if inst.is_connected_ws:

                    self.ws_facade = CoinigyFacade(inst, api_context)
                    if not self.possible_channels:

                        def pop_chans(chans):
                            """Populate the channels on success"""
                            self.possible_channels = chans

                        self.ws_facade.get_channels(pop_chans)

                elif inst.is_connected_ws is not None:
                    self.api_facade = \
                        CoinigyFacade(inst, api_context)

        parser = CoinigyResultParserFactory(self._strategy_data, context)
        parser.product.interface()

        # initialze supplemented data
        context["strategy"].update({
            "coinigy": {
                "data": parser.product.result,
                "api": self.api_facade,
                "ws": self.ws_facade,
                }
            })

        return context



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'marshmallow',
    'requests',
    'urllib3',
    'bors',
    'click',
]

test_requirements = [
    'pytest',
    'tox',
    'coverage',
    'pytest-cov',
]


setup_requirements = ['pytest-runner', ]

setup(
    author="Bobby",
    author_email='karma0@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="A cryptocurrency trading bot",
    entry_points={
        'console_scripts': [
            'nombot=nombot.cli:main',
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='nombot cryptocurrency bot',
    name='nombot',
    packages=find_packages(exclude=['tests', 'docs']),
    package_dir={'nombot': 'nombot'},
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/karma0/nombot',
    version='2.2.7',
    zip_safe=False,
)
